# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
from .. import plotlib as pplt
import numpy as np
import unum
import copy
from scipy import stats
from scipy import ndimage
from ..utils.types import ARRAYTYPES, INTEGERTYPES, NUMBERTYPES, STRINGTYPES
from ..utils import make_unit


class Points(object):

    def __init__(self, xy=np.empty((0, 2), dtype=float), v=[],
                 unit_x='', unit_y='', unit_v='', name=''):
        """
        Class representing a set of points.
        You can use 'make_unit' to provide unities.

        Parameters
        ----------
        xy : nx2 array.
            Representing the coordinates of each point of the set (n points).
        v : n array, optional
            Representing values attached at each points.
        unit_x : Unit object, optional
            X unit_y.
        unit_y : Unit object, optional
            Y unit_y.
        unit_v : Unit object, optional
            values unit_y.
        name : string, optional
            Name of the points set
        """
        self.__v = []
        if len(xy) == 0:
            xy = np.empty((0, 2), dtype=float)
        self.xy = xy
        self.v = v
        self.unit_v = unit_v
        self.unit_x = unit_x
        self.unit_y = unit_y
        self.name = name

    def __iter__(self):
        if self.v is None or len(self.v) == 0:
            for i in np.arange(len(self.xy)):
                yield self.xy[i]
        else:
            for i in np.arange(len(self.xy)):
                yield self.xy[i], self.v[i]

    def __len__(self):
        return self.xy.shape[0]

    def __add__(self, another):
        if isinstance(another, Points):
            # trivial additions
            if len(self.xy) == 0:
                return another.copy()
            elif len(another.xy) == 0:
                return self.copy()
            # checking unit systems
            if len(self.xy) != 0:
                try:
                    self.unit_x + another.unit_x
                    self.unit_y + another.unit_y
                    if self.v is not None and another.v is not None:
                        self.unit_v + another.unit_v
                except unum.IncompatibleUnitsError:
                    raise ValueError("Units system are not the same")
            else:
                self.unit_x = another.unit_x
                self.unit_y = another.unit_y
                if another.v is not None:
                    self.unit_v = another.unit_v
            # compacting coordinates
            if another.xy.shape == (0,):
                new_xy = self.xy
            elif self.xy.shape == (0,):
                xy = another.xy
                xy[:, 0] = xy[:, 0]*(self.unit_x/another.unit_x).asNumber()
                xy[:, 1] = xy[:, 1]*(self.unit_y/another.unit_y).asNumber()
                new_xy = xy
            elif another.xy.shape == (0,) and self.xy.shape == (0,):
                new_xy = np.array([[]])
            else:
                xy = another.xy
                xy[:, 0] = xy[:, 0]*(self.unit_x/another.unit_x).asNumber()
                xy[:, 1] = xy[:, 1]*(self.unit_y/another.unit_y).asNumber()
                new_xy = np.append(self.xy, xy, axis=0)
            # testing v presence
            v_presence = True
            if self.v is None and another.v is None:
                if len(self.xy) != 0:
                    v_presence = False
            elif self.v is not None and another.v is not None:
                    v_presence = True
            else:
                raise Exception()
            # compacting points and returning
            if v_presence:
                if self.v is None and another.v is None:
                    v = np.array([])
                elif self.v is None:
                    v = another.v*(self.unit_v/another.unit_v).asNumber()
                elif another.v is None:
                    v = self.v
                else:
                    v_tmp = another.v*(self.unit_v/another.unit_v).asNumber()
                    v = np.append(self.v, v_tmp)
                return Points(new_xy, v,
                              unit_x=self.unit_x,
                              unit_y=self.unit_y,
                              unit_v=self.unit_v)
            else:
                return Points(new_xy,
                              unit_x=self.unit_x,
                              unit_y=self.unit_y)
        else:
            raise Exception("You can't add {} to Points objects"
                            .format(type(another)))

    def __eq__(self, obj):
        if not isinstance(obj, Points):
            return False
        if not np.all(self.xy == obj.xy):
            return False
        if not np.all(self.v == obj.v):
            return False
        if not self.unit_x == obj.unit_x:
            return False
        if not self.unit_y == obj.unit_y:
            return False
        if not self.unit_v == obj.unit_v:
            return False
        return True

    @property
    def xy(self):
        return self.__xy

    @xy.setter
    def xy(self, values):
        values = np.asarray(values, dtype=float)
        if len(values != 0):
            if not values.ndim == 2:
                raise ValueError("ndim of xy is {} and should be 2"
                                 .format(values.ndim))
            if not values.shape[1] == 2:
                raise ValueError()
        self.__xy = values
        if len(values) != len(self.__v):
            self.__v = np.array([])

    @xy.deleter
    def xy(self):
        raise Exception("Nope, can't do that")

    @property
    def v(self):
        return self.__v

    @v.setter
    def v(self, values):
        if values is None:
            self.__v = None
        else:
            values = np.array(values, subok=True)
            if not values.ndim == 1:
                raise ValueError()
            if not len(values) in [0, len(self.__xy)]:
                raise ValueError()
            self.__v = values

    @v.deleter
    def v(self):
        raise Exception("Nope, can't do that")

    @property
    def unit_x(self):
        return self.__unit_x

    @unit_x.setter
    def unit_x(self, unit):
        if isinstance(unit, unum.Unum):
            self.__unit_x = unit
        elif isinstance(unit, STRINGTYPES):
            try:
                self.__unit_x = make_unit(unit)
            except (ValueError, TypeError):
                raise Exception()
        else:
            raise Exception()

    @unit_x.deleter
    def unit_x(self):
        raise Exception("Nope, can't delete 'unit_x'")

    @property
    def unit_y(self):
        return self.__unit_y

    @unit_y.setter
    def unit_y(self, unit):
        if isinstance(unit, unum.Unum):
            self.__unit_y = unit
        elif isinstance(unit, STRINGTYPES):
            try:
                self.__unit_y = make_unit(unit)
            except (ValueError, TypeError):
                raise Exception()
        else:
            raise Exception()

    @unit_y.deleter
    def unit_y(self):
        raise Exception("Nope, can't delete 'unit_y'")

    @property
    def unit_v(self):
        return self.__unit_v

    @unit_v.setter
    def unit_v(self, unit):
        if isinstance(unit, unum.Unum):
            self.__unit_v = unit
        elif isinstance(unit, STRINGTYPES):
            try:
                self.__unit_v = make_unit(unit)
            except (ValueError, TypeError):
                raise Exception()
        else:
            try:
                self.__unit_v = make_unit(str(unit))
            except Exception as msg:
                print(msg)
                print(type(unit))
                print(unit)
                raise Exception()

    @unit_v.deleter
    def unit_v(self):
        raise Exception("Nope, can't delete 'unit_v'")

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        if isinstance(name, STRINGTYPES):
            self.__name = name
        else:
            raise Exception()

    @name.deleter
    def name(self):
        raise Exception("Nope, can't delete 'name'")

    def copy(self):
        """
        Return a copy of the Points object.
        """
        return copy.deepcopy(self)

    def get_points_density(self, bw_method=None, resolution=100,
                           output_format='normalized', raw=False):
        """
        Return a ScalarField with points density.

        Parameters
        ----------
        bw_method : str, scalar or callable, optional
            The method used to calculate the estimator bandwidth.
            This can be 'scott', 'silverman', a scalar constant or
            a callable. If a scalar, this will be used as percent of
            the data std. If a callable, it should take a gaussian_kde
            instance as only parameter and return a scalar.
            If None (default), 'scott' is used.
        resolution : integer, optional
            Resolution for the resulting field.
        output_format : string, optional
            'normalized' (default) : give position probability
                                     (integral egal 1).
            'ponderated' : give position probability ponderated by the number
                           or points (integral egal number of points).
            'concentration' : give local concentration (in point per surface).
        raw : boolean, optional
            If 'False' (default), return a ScalarField object,
            if 'True', return numpy array.

        Returns
        -------
        density : array, ScalarField object or None
            Return 'None' if there is not enough points in the cloud.
        """
        # checking points length
        if len(self.xy) < 2:
            return None
        # getting data
        min_x = np.min(self.xy[:, 0])
        max_x = np.max(self.xy[:, 0])
        min_y = np.min(self.xy[:, 1])
        max_y = np.max(self.xy[:, 1])
        # getting min, max values and resolution in each direction
        width_x = max_x - min_x
        width_y = max_y - min_y
        if width_x == 0 and width_y == 0:
            raise Exception()
        elif width_x == 0:
            min_x = min_x - width_y/2.
            max_x = max_x + width_y/2.
            width_x = width_y
            res_x = resolution
            res_y = resolution
        elif width_y == 0:
            min_y = min_y - width_x/2.
            max_y = max_y + width_x/2.
            width_y = width_x
            res_x = resolution
            res_y = resolution
        elif width_x > width_y:
            res_x = resolution
            res_y = int(np.round(resolution*width_y/width_x))
        else:
            res_y = resolution
            res_x = int(np.round(resolution*width_x/width_y))
        if res_x < 2 or res_y < 2:
            raise ValueError()
        # check potential singular covariance matrix situations
        if (np.all(self.xy[:, 0] == self.xy[0, 0]) or
                np.all(self.xy[:, 1] == self.xy[0, 1])):
            return None
        # get kernel using scipy
        std = np.mean([np.std(self.xy[:, 0]), np.std(self.xy[:, 1])])
        if isinstance(bw_method, NUMBERTYPES):
            if width_x > width_y:
                ad_len = width_y
            else:
                ad_len = width_x
            ad_bw_method = bw_method*std/ad_len
        else:
            ad_bw_method = bw_method
        kernel = stats.gaussian_kde(self.xy.transpose(),
                                    bw_method=ad_bw_method)
        # little adaptation to avoid streched density map
        if width_x > width_y:
            kernel.inv_cov[0, 0] = np.max([kernel.inv_cov])
        else:
            kernel.inv_cov[1, 1] = np.max([kernel.inv_cov])
        kernel.inv_cov[0, 1] *= 0
        kernel.inv_cov[1, 0] *= 0
        # creating grid
        if width_x > width_y:
            dx_border = kernel.factor*width_y/2.
            dy_border = dx_border
        else:
            dx_border = kernel.factor*width_x/2.
            dy_border = dx_border
        axe_x = np.linspace(min_x - dx_border, max_x + dx_border, res_x)
        axe_y = np.linspace(min_y - dy_border, max_y + dy_border, res_y)
        X, Y = np.meshgrid(axe_x, axe_y)
        X = X.flatten()
        Y = Y.flatten()
        positions = np.array([[X[i], Y[i]]
                              for i in np.arange(len(X))]).transpose()
        # estimating density
        values = kernel(positions)
        values = values.reshape((res_y, res_x)).transpose()
        # normalize (not normalized yet because of the modification of inv_cov)
        dx = axe_x[1] - axe_x[0]
        dy = axe_y[1] - axe_y[0]
        values /= np.sum(np.sum(values))*(dx)*(dy)
        # adapt to wanted output_format
        if output_format is None or output_format == "normalized":
            unit_values = make_unit('')
        elif output_format == 'ponderated':
            values = values*len(self.xy)
            unit_values = make_unit('')
        elif output_format == "percentage":
            values = values*100
            unit_values = make_unit('')
        elif output_format == "concentration":
            unit_values = 1/self.unit_x/self.unit_y
            values = values*len(self.xy)
        else:
            raise ValueError()
        # return
        if np.all(np.isnan(values)) or np.all(values == np.inf):
            return None
        if raw:
            return values
        else:
            from .scalarfield import ScalarField
            sf = ScalarField()
            sf.import_from_arrays(axe_x, axe_y, values, mask=False,
                                  unit_x=self.unit_x, unit_y=self.unit_y,
                                  unit_values=unit_values)
            return sf

    def get_points_density2(self, res, subres=None, raw=False,
                            ponderated=False):
        """
        Return a ScalarField with points density.

        Parameters
        ----------
        res : number or 2x1 array of numbers
            fdensity field number of subdivision.
            Can be the same number for both axis,  or one number per axis
            (need to give a tuple).
        raw : boolean, optional
            If 'False' (default), return a ScalarField object,
            if 'True', return numpy array.
        ponderated : boolean, optiona
            If 'True', values associated to points are used to ponderate the
            density field. Default is 'False'.
        subres : odd integer, optional
            If specified, a subgrid of resolution res*subres is used to
            make result more accurate.
        """
        # checking parameters
        if isinstance(res, int):
            res_x = res
            res_y = res
        elif isinstance(res, ARRAYTYPES):
            if len(res) != 2:
                raise ValueError()
            res_x = res[0]
            res_y = res[1]
        else:
            raise TypeError()
        if not isinstance(raw, bool):
            raise TypeError()
        if not isinstance(ponderated, bool):
            raise TypeError()
        if isinstance(subres, int) and subres > 0:
            subres = np.floor(subres/2)*2
            subres2 = (subres)/2
        elif subres is None:
            pass
        else:
            raise TypeError()
        # If we use a subgrid
        if subres is not None:
            # creating grid
            min_x = np.min(self.xy[:, 0])
            max_x = np.max(self.xy[:, 0])
            min_y = np.min(self.xy[:, 1])
            max_y = np.max(self.xy[:, 1])
            dx = (max_x - min_x)/(res_x)
            dy = (max_y - min_y)/(res_y)
            sub_dx = dx/subres
            sub_dy = dy/subres
            axe_x = np.arange(min_x - dx/2, max_x + dx/2 + sub_dx, sub_dx)
            axe_y = np.arange(min_y - dy/2, max_y + dy/2 + sub_dy, sub_dy)
            values = np.zeros((len(axe_x), len(axe_y)))
            # filling grid with density
            for i, pt in enumerate(self.xy):
                x = pt[0]
                y = pt[1]
                ind_x = np.argmin(np.abs(axe_x - x))
                ind_y = np.argmin(np.abs(axe_y - y))
                slic_x = slice(ind_x - subres2 + 1, ind_x + subres2)
                slic_y = slice(ind_y - subres2 + 1, ind_y + subres2)
                if ponderated:
                    values[slic_x, slic_y] += self.v[i]
                else:
                    values[slic_x, slic_y] += 1
            values /= (dx*dy)
            values = values[subres2:-subres2, subres2:-subres2]
            axe_x = axe_x[subres2:-subres2]
            axe_y = axe_y[subres2:-subres2]
        # if we do not use a subgrid
        else:
            # creating grid
            min_x = np.min(self.xy[:, 0])
            max_x = np.max(self.xy[:, 0])
            min_y = np.min(self.xy[:, 1])
            max_y = np.max(self.xy[:, 1])
            axe_x, dx = np.linspace(min_x, max_x, res_x, retstep=True)
            axe_y, dy = np.linspace(min_y, max_y, res_y, retstep=True)
            values = np.zeros((len(axe_x), len(axe_y)))
            # filling grid with density
            for i, pt in enumerate(self.xy):
                x = pt[0]
                y = pt[1]
                ind_x = np.argmin(np.abs(axe_x - x))
                ind_y = np.argmin(np.abs(axe_y - y))
                if ponderated:
                    values[ind_x, ind_y] += self.v[i]
                else:
                    values[ind_x, ind_y] += 1
            values /= (dx*dy)
        # return the field
        if raw:
            return values
        else:
            from .scalarfield import ScalarField
            sf = ScalarField()
            if ponderated:
                unit_values = self.unit_v/self.unit_x/self.unit_y
            else:
                unit_values = 1/self.unit_x/self.unit_y
            sf.import_from_arrays(axe_x, axe_y, values, mask=False,
                                  unit_x=self.unit_x, unit_y=self.unit_y,
                                  unit_values=unit_values)
            return sf

    def get_clusters(self, eps, min_samples=5):
        """
        Perform DBSCAN clustering from vector array or distance matrix.
        (see sklearn.cluster.DBSCAN)

        Notes
        ------
        DBSCAN - Density-Based Spatial Clustering of Applications with Noise.
        Finds core samples of high density and expands clusters from them.
        Good for data which contains clusters of similar density.
        """
        try:
            import sklearn.cluster as clst
        except ImportError:
            raise Exception("You need to install `sklearn` to use this "
                            "functionnality")
        X = self.xy
        db = clst.DBSCAN(eps=eps, min_samples=min_samples).fit(X)
        labels = db.labels_
        uniq_labels = np.array(list(set(labels)))
        nmb_cluster = len(uniq_labels)
        if -1 in uniq_labels:
            nmb_cluster -= 1
        # create separated points objects
        points = []
        for lab in uniq_labels:
            filt = labels == lab
            if len(self.v) == len(self.xy):
                new_v = self.v[filt]
            else:
                new_v = self.v
            tmp_pts = Points(self.xy[filt], new_v, unit_x=self.unit_x,
                             unit_y=self.unit_y, unit_v=self.unit_v)
            points.append(tmp_pts)
        return points

    def get_envelope(self, alpha=None):
        """
        Return the convex or concave hull (if alpha specified) for the set of
        points.

        Parameters
        ----------
        alpha : number
            maximum distance between two points of the hull.

        Notes
        -----
        Credit to mlaloux
        (https://github.com/mlaloux/Python--alpha-shape_concave_hull)
        """
        # import shapely functions
        try:
            from shapely.geometry import mapping
            from shapely.geometry import MultiLineString
            from shapely.ops import polygonize, cascaded_union
        except ImportError:
            raise Exception("This functionnality need 'shapely' module")
        if alpha is None:
            alpha = np.inf
        # triangulate
        from scipy.spatial import Delaunay
        points = self.xy
        tri = Delaunay(points)
        # add and sort points
        edges = set()
        edge_points = []

        def add_edge(i, j):
            """Add a line between the i-th and j-th points,
            if not in the list already"""
            if (i, j) in edges or (j, i) in edges:
                return
            edges.add((i, j))
            edge_points.append(points[[i, j]])
        points = np.array(points)
        for ia, ib, ic in tri.vertices:
            pa = points[ia]
            pb = points[ib]
            pc = points[ic]
            # Lengths of sides of triangle
            a = np.sqrt((pa[0]-pb[0])**2 + (pa[1]-pb[1])**2)
            b = np.sqrt((pb[0]-pc[0])**2 + (pb[1]-pc[1])**2)
            c = np.sqrt((pc[0]-pa[0])**2 + (pc[1]-pa[1])**2)
            # Semiperimeter of triangle
            s = (a + b + c)/2.0
            # Area of triangle by Heron's formula
            area = np.sqrt(s*(s-a)*(s-b)*(s-c))
            circum_r = a*b*c/(4.0*area)
            # Here's the radius filter.
            #if circum_r < 1.0/alpha:
            if circum_r < alpha:
                add_edge(ia, ib)
                add_edge(ib, ic)
                add_edge(ic, ia)
        # concatenate polygons
        m = MultiLineString(edge_points)
        triangles = list(polygonize(m))
        hull = mapping(cascaded_union(triangles))['coordinates'][0]
        hull = np.array(hull, dtype=float)
        if hull.ndim == 3:
            hull = hull[0]
        # transform to Points object
        pt = Points(xy=hull, v=[], unit_x=self.unit_x,
                    unit_y=self.unit_y)
        return pt

    def get_velocity(self, incr=1, smooth=0, xaxis='time'):
        """
        Assuming that associated 'v' values are times for each points,
        compute the velocity of the trajectory.

        Parameters
        ----------
        incr : integer, optional
            Increment use to get used points (default is 1).
        smooth : number, optional
            Cut off frequency for the lowpass filter.
        xaxis : string, optional
            Value to put in the profile x axis, can be 'time' (default), 'x'
            or 'y'.

        Returns
        -------
        Vx : Profile object
            Profile of x velocity versus time.
        Vy : Profile object
            Profile of y velocity versus time.
        """
        if smooth < 0:
            raise ValueError()
        if xaxis not in ['time', 'x', 'y']:
            raise ValueError()
        # checking 'v' presence
        if len(self.v) == 0:
            raise Exception()
        # sorting points by time
        ind_sort = np.argsort(self.v)
        times = self.v[ind_sort]
        xy = self.xy[ind_sort]
        x = xy[:, 0]
        y = xy[:, 1]
        # using increment if necessary
        if incr != 1:
            x = x[::incr]
            y = y[::incr]
            times = times[::incr]
        dx = x[1::] - x[:-1]
        dy = y[1::] - y[:-1]
        dt = times[1::] - times[:-1]
        # smoothing if necessary
        if smooth != 0:
            tmp_pts = Points(list(zip(x, y)), times)
            tmp_pts.smooth(tos='lowpass', size=smooth, inplace=True)
            x = tmp_pts.xy[:, 0]
            y = tmp_pts.xy[:, 1]
        # getting velocity between points
        Vx = np.array([(x[i + 1] - x[i])/dt[i]
                       for i in np.arange(len(x) - 1)])
        Vy = np.array([(y[i + 1] - y[i])/dt[i]
                       for i in np.arange(len(y) - 1)])
        # getting xaxis
        if xaxis == 'time':
            x_prof = times[:-1] + dt/2.
        elif xaxis == 'x':
            x_prof = x[:-1] + dx/2.
        elif xaxis == 'y':
            x_prof = y[:-1] + dy/2.
        # returning profiles
        from .profile import Profile
        unit_Vx = self.unit_x/self.unit_v
        Vx *= unit_Vx.asNumber()
        unit_Vx /= unit_Vx.asNumber()
        prof_x = Profile(x_prof, Vx, mask=False, unit_x=self.unit_v,
                         unit_y=unit_Vx)
        unit_Vy = self.unit_y/self.unit_v
        Vy *= unit_Vy.asNumber()
        unit_Vy /= unit_Vy.asNumber()
        prof_y = Profile(x_prof, Vy, mask=False, unit_x=self.unit_v,
                         unit_y=unit_Vy)
        return prof_x, prof_y

    def get_evolution_on_sf(self, SF, axe_x=None):
        """
        Return the evolution of the value represented by a scalar field, on
        the path of the trajectory.

        Parameters
        ----------
        SF : ScalarField object
        axe_x : string, optional
            What put in the x axis (can be 'x', 'y', 'v').
            default is 'v' when available and 'x' else.

        Returns
        -------
        evol : Profile object
        """
        # check parameters
        from .scalarfield import ScalarField
        if not isinstance(SF, ScalarField):
            raise TypeError()
        if len(self.xy) == 0:
            from .profile import Profile
            return Profile()
        if axe_x is None:
            if len(self.v) == len(self.xy):
                axe_x = 'v'
            else:
                axe_x = 'x'
        if not isinstance(axe_x, STRINGTYPES):
            raise TypeError()
        # get x values
        if axe_x == 'v':
            if len(self.v) == 0:
                raise ValueError()
            x_prof = self.v
            unit_x = self.unit_v
        elif axe_x == 'x':
            x_prof = self.xy[:, 0]
            unit_x = self.unit_x
        elif axe_x == 'y':
            x_prof = self.xy[:, 1]
            unit_x = self.unit_y
        else:
            raise ValueError()
        # get the y value
        y_prof = np.empty((len(self.xy)), dtype=float)
        for i, pt in enumerate(self.xy):
            y_prof[i] = SF.get_value(*pt, ind=False, unit=False)
        mask = np.isnan(y_prof)
        unit_y = SF.unit_values
        # returning
        evol = Profile(x_prof, y_prof, mask=mask, unit_x=unit_x,
                       unit_y=unit_y)
        return evol

    def get_evolution_on_tsf(self, TSF, axe_x=None):
        """
        Return the evolution of the value represented by scalar fields, on
        the path of the trajectory.
        Timse of the TSF must be consistent with the times of the Points.

        Parameters
        ----------
        TSF : tsf.TemporalScalarField object
        axe_x : string, optional
            What put in the x axis (can be 'x', 'y', 'v').
            default is 'v' (associated with time)

        Returns
        -------
        evol : Profile object
        """
        # check parameters
        from . import temporalscalarfields as tsf
        from .profile import Profile
        if not isinstance(TSF, tsf.TemporalScalarFields):
            raise TypeError()
        if len(self.xy) == 0:
            return Profile()
        if axe_x is None:
            axe_x = 'v'
        if not isinstance(axe_x, STRINGTYPES):
            raise TypeError()
        # get x values
        if axe_x == 'v':
            if len(self.v) == 0:
                raise ValueError()
            x_prof = self.v
            unit_x = self.unit_v
        elif axe_x == 'x':
            x_prof = self.xy[:, 0]
            unit_x = self.unit_x
        elif axe_x == 'y':
            x_prof = self.xy[:, 1]
            unit_x = self.unit_y
        else:
            raise ValueError()
        # get the y value
        times = self.v
        y_prof = np.empty((len(self.xy)), dtype=float)
        for i, pt in enumerate(self.xy):
            time = times[i]
            SF = TSF.fields[TSF.times == time][0]
            y_prof[i] = SF.get_value(*pt, ind=False, unit=False)
        mask = np.isnan(y_prof)
        unit_y = TSF.unit_values
        # returning
        evol = Profile(x_prof, y_prof, mask=mask, unit_x=unit_x,
                       unit_y=unit_y)
        return evol

    def fit(self, kind='polynomial', order=2, simplify=False):
        """
        Return the parametric coefficients of the fitting curve on the points.

        Parameters
        ----------
        kind : string, optional
            The kind of fitting used. Can be 'polynomial' or 'ellipse'.
        order : integer
            Approximation order for the fitting.
        Simplify : boolean or string, optional
            Can be False (default), 'x' or 'y'. Perform a simplification
            (see Points.Siplify()) before the fitting.

        Returns
        -------
        p : array, only for polynomial fitting
            Polynomial coefficients, highest power first
        radii : array, only for ellipse fitting
            Ellipse demi-axes radii.
        center : array, only for ellipse fitting
           Ellipse center coordinates.
        alpha : number
            Angle between the x axis and the major axis.
        """
        if not isinstance(order, int):
            raise TypeError("'order' must be an integer")
        if not isinstance(kind, STRINGTYPES):
            raise TypeError("'kind' must be a string")
        if not simplify:
            xytmp = self.xy
        elif simplify == 'x':
            xytmp = self.simplify(axe=0).xy
        elif simplify == 'y':
            xytmp = self.simplify(axe=1).xy

        if kind == 'polynomial':
            p = np.polyfit(xytmp[:, 0], xytmp[:, 1], deg=order)
            return p
        elif kind == 'ellipse':
            import fit_ellipse as fte
            res = fte.fit_ellipse(xytmp)
            radii, center, alpha = fte.get_parameters(res)
            return radii, center, alpha

    def add(self, pt, v=None):
        """
        Add a new point.

        Parameters
        ----------
        pt : 2x1 array of numbers
            Point to add.
        v : number, optional
            Value of the point (needed if other points have values).
        """
        # check parameters
        if not isinstance(pt, ARRAYTYPES):
            raise TypeError()
        pt = np.array(pt, subok=True)
        if not pt.shape == (2,):
            raise ValueError()
        if not isinstance(pt[0], NUMBERTYPES):
            raise TypeError()
        if v is not None:
            if not isinstance(v, NUMBERTYPES):
                raise TypeError()
        # store new data
        self.__xy = np.append(self.xy, [pt], axis=0)
        if v is None and self.v.shape[0] != 0:
            raise ValueError('You should specify an associated value : v')
        if v is not None:
            self.__v = np.append(self.v, v)

    def remove(self, ind):
        """
        Remove the point number 'ind' of the points cloud.
        In place.

        Parameters
        ----------
        ind : integer or array of integer
        """
        if isinstance(ind, INTEGERTYPES):
            ind = [ind]
        elif isinstance(ind, ARRAYTYPES):
            if not np.all([isinstance(val, int) for val in ind]):
                raise TypeError("'ind' must be an integer or an array of"
                                " integer")
            ind = np.array(ind)
        else:
            raise TypeError("'ind' must be an integer or an array of integer")
        tmp_v = self.v.copy()
        self.xy = np.delete(self.xy, ind, axis=0)
        if len(self.v) == len(self.xy) + 1:
            self.v = np.delete(tmp_v, ind, axis=0)

    def change_unit(self, axe, new_unit):
        """
        Change the unit of an axe.

        Parameters
        ----------
        axe : string
            'y' for changing the profile y axis unit
            'x' for changing the profile x axis unit
            'v' for changing the profile values unit
        new_unit : Unum.unit object or string
            The new unit.
        """
        if isinstance(new_unit, STRINGTYPES):
            new_unit = make_unit(new_unit)
        if not isinstance(new_unit, unum.Unum):
            raise TypeError()
        if not isinstance(axe, STRINGTYPES):
            raise TypeError()
        if axe == 'x':
            old_unit = self.unit_x
            new_unit = old_unit.asUnit(new_unit)
            fact = new_unit.asNumber()
            self.xy[:, 0] *= fact
            self.unit_x = new_unit/fact
        elif axe == 'y':
            old_unit = self.unit_y
            new_unit = old_unit.asUnit(new_unit)
            fact = new_unit.asNumber()
            self.xy[:, 1] *= fact
            self.unit_y = new_unit/fact
        elif axe == 'v':
            old_unit = self.unit_v
            new_unit = old_unit.asUnit(new_unit)
            fact = new_unit.asNumber()
            self.v *= fact
            self.unit_v = new_unit/fact
        else:
            raise ValueError()

    def set_origin(self, x=None, y=None):
        """
        Set the given point (x, y) as the new referential.

        Parameters
        ----------
        x: number
            .
        y: number
            .
        """
        # Check
        if x is not None:
            if not isinstance(x, NUMBERTYPES):
                raise TypeError()
            self.xy[:, 0] -= x
        if y is not None:
            if not isinstance(y, NUMBERTYPES):
                raise TypeError()
            self.xy[:, 1] -= y

    def crop(self, intervx=None, intervy=None, intervv=None, inplace=True,
             ind=False):
        """
        Crop the points cloud.

        Parameters
        ----------
        intervx : 2x1 tuple
            Interval on x axis
        intervy : 2x1 tuple
            Interval on y axis
        intervv : 2x1 tuple
            Interval on v values

        Returns
        -------
        tmp_pts : Points object
            croped version of the point cloud.
        """
        #
        if inplace:
            tmp_pts = self
        else:
            tmp_pts = self.copy()
        # Getting cropping mask
        if ind:
            if intervx is not None or intervy is not None:
                raise ValueError()
            mask = np.ones(len(self.xy), dtype=bool)
            mask[intervv[0]:intervv[1]] = False
        else:
            mask = np.zeros(len(self.xy), dtype=bool)
            if intervx is not None:
                out_zone = np.logical_or(self.xy[:, 0] < intervx[0],
                                         self.xy[:, 0] > intervx[1])
                mask = np.logical_or(mask, out_zone)
            if intervy is not None:
                out_zone = np.logical_or(self.xy[:, 1] < intervy[0],
                                         self.xy[:, 1] > intervy[1])
                mask = np.logical_or(mask, out_zone)
            if intervv is not None and len(self.v) != 0:
                out_zone = np.logical_or(self.v < intervv[0],
                                         self.v > intervv[1])
                mask = np.logical_or(mask, out_zone)
        # Cropping values
        tmp_pts.__xy = tmp_pts.xy[~mask, :]
        if len(tmp_pts.v) != 0:
            tmp_pts.__v = tmp_pts.v[~mask]
        # returning
        if not inplace:
            return tmp_pts

    def cut(self, intervx=None, intervy=None):
        """
        Return a point cloud where the given area has been removed.

        Parameters
        ----------
        intervx : 2x1 tuple
            Interval on x axis
        intervy : 2x1 tuple
            Interval on y axis

        Returns
        -------
        tmp_pts : Points object
            Cutted version of the point cloud.
        """
        tmp_pts = self.copy()
        mask = np.ones(len(self.xy))
        if intervx is not None:
            out_zone = np.logical_and(self.xy[:, 0] > intervx[0],
                                      self.xy[:, 0] < intervx[1])
            mask = np.logical_and(mask, out_zone)
        if intervy is not None:
            out_zone = np.logical_and(self.xy[:, 1] > intervy[0],
                                      self.xy[:, 1] < intervy[1])
            mask = np.logical_and(mask, out_zone)
        tmp_pts.xy = tmp_pts.xy[~mask, :]
        if len(tmp_pts.v) != 0:
            tmp_pts.v = tmp_pts.v[~mask]
        return tmp_pts

    def scale(self, scalex=1., scaley=1., scalev=1., inplace=False):
        """
        Change the scale of the axis.

        Parameters
        ----------
        scalex, scaley, scalev : numbers or Unum objects
            scales along x, y and v
        inplace : boolean, optional
            If 'True', scaling is done in place, else, a new instance is
            returned.
        """
        # check params
        if not isinstance(scalex, NUMBERTYPES + (unum.Unum, )):
            raise TypeError()
        if not isinstance(scaley, NUMBERTYPES + (unum.Unum, )):
            raise TypeError()
        if not isinstance(scalev, NUMBERTYPES + (unum.Unum, )):
            raise TypeError()
        if not isinstance(inplace, bool):
            raise TypeError()
        if inplace:
            tmp_pt = self
        else:
            tmp_pt = self.copy()
        # adapt unit
        if isinstance(scalex, unum.Unum):
            new_unit = scalex*tmp_pt.unit_x
            fact = new_unit.asNumber()
            new_unit /= fact
            tmp_pt.unit_x = new_unit
            scalex = fact
        if isinstance(scaley, unum.Unum):
            new_unit = scaley*tmp_pt.unit_y
            fact = new_unit.asNumber()
            new_unit /= fact
            tmp_pt.unit_y = new_unit
            scaley = fact
        if isinstance(scalev, unum.Unum):
            new_unit = scalev*tmp_pt.unit_v
            fact = new_unit.asNumber()
            new_unit /= fact
            tmp_pt.unit_v = new_unit
            scalev = fact
        # loop
        if scalex != 1. or scaley != 1.:
            tmp_pt.xy *= np.array([scalex, scaley])
        if scalev != 1.:
            tmp_pt.v *= scalev
        # returning
        if not inplace:
            return tmp_pt

    def rotate(self, angle, inplace=False):
        """
        Rotate the point set.

        Parameters
        ----------
        angle: number
            Rotation angle in radian.
        """
        # Checks
        if inplace:
            tmpp = self
        else:
            tmpp = self.copy()
        #
        x, y = self.xy[:, 0], self.xy[:, 1]
        new_x = x*np.cos(angle) - y*np.sin(angle)
        new_y = y*np.cos(angle) + x*np.sin(angle)
        tmpp.xy[:, 0] = new_x
        tmpp.xy[:, 1] = new_y
        # Return
        return tmpp


    def remove_doublons(self, method='average', inplace=False, eps_rel=1e-6):
        """
        Replace values associated to the same 'v' by their average.

        Parameters
        ----------
        method : string in {'average', 'max', 'min'}
           Method used to remove the doublons.
        """
        if len(self.v) == 0:
            raise Exception()
        if inplace:
            tmp_pt = self
        else:
            tmp_pt = self.copy()
        vs = tmp_pt.v
        ord_magn = (np.sum(vs**2)/len(vs))**.5
        nmb_dec = -int(round(np.log10(ord_magn*eps_rel)))
        tmp_vs = np.round(vs, decimals=nmb_dec)
        new_vs = np.sort(list(set(tmp_vs)))
        if method == 'average':
            new_xy = [np.mean(tmp_pt.xy[tmp_vs == vi], axis=0)
                      for vi in new_vs]
        elif method == 'min':
            new_xy = [np.min(tmp_pt.xy[tmp_vs == vi], axis=0)
                      for vi in new_vs]
        elif method == 'max':
            new_xy = [np.max(tmp_pt.xy[tmp_vs == vi], axis=0)
                      for vi in new_vs]
        else:
            raise ValueError()
        tmp_pt.xy = new_xy
        tmp_pt.v = new_vs
        # returning
        if not inplace:
            return tmp_pt

    def reverse(self):
        """
        Return a Points object where x and y axis are swaped.
        """
        tmp_pt = self.copy()
        xy_tmp = tmp_pt.xy*0
        xy_tmp[:, 0] = tmp_pt.xy[:, 1]
        xy_tmp[:, 1] = tmp_pt.xy[:, 0]
        tmp_pt.xy = xy_tmp
        return tmp_pt

    def decompose(self):
        """
        return a tuple of Points object, with only one point per object.
        """
        if len(self) == 1:
            return [self]
        if len(self) != len(self.v):
            raise Exception()
        pts_tupl = []
        for i in np.arange(len(self)):
            pts_tupl.append(Points([self.xy[i]], [self.v[i]], self.unit_x,
                                   self.unit_y, self.unit_v, self.name))
        return pts_tupl

    def sort(self, ref='x', inplace=False):
        """
        Sort the points according to the reference.

        Parameters
        ----------
        ref : string or array of indice
            can be 'x', 'y' or 'v' to sort according those values or an
            array of indice
        inplace ; boolean
            If 'True', sort in place, else, return an new sorted instance.
        """
        # check parameters
        if isinstance(ref, STRINGTYPES):
            if ref not in ['x', 'y', 'v']:
                raise ValueError
        elif isinstance(ref, ARRAYTYPES):
            ref = np.array(ref, dtype=int)
            if len(ref) != len(self.xy):
                raise ValueError()
        else:
            raise TypeError()
        if not isinstance(inplace, bool):
            raise TypeError()
        # get order
        if ref == 'x':
            order = np.argsort(self.xy[:, 0])
        elif ref == 'y':
            order = np.argsort(self.xy[:, 1])
        elif ref == 'v':
            if len(self.v) == 0:
                raise ValueError()
            order = np.argsort(self.v)
        else:
            order = ref
        # reordering
        if inplace:
            tmp_pt = self
        else:
            tmp_pt = self.copy()
        tmp_pt.xy = tmp_pt.xy[order]
        if len(tmp_pt.v) == len(tmp_pt.xy):
            tmp_pt.v = tmp_pt.v[order]
        # returning
        if not inplace:
            return tmp_pt

    def order_on_line(self, inplace=False):
        """
        Re-order the set of points to try to create a line.
        """
        if inplace:
            tmp_pts = self
        else:
            tmp_pts = self.copy()
        #
        from sklearn.neighbors import NearestNeighbors
        clf = NearestNeighbors(2).fit(self.xy)
        G = clf.kneighbors_graph()
        import networkx as nx
        T = nx.from_scipy_sparse_matrix(G)
        order = list(nx.dfs_preorder_nodes(T, 0))
        paths = [list(nx.dfs_preorder_nodes(T, i)) for i in range(len(self.xy))]
        mindist = np.inf
        minidx = 0
        for i in range(len(self.xy)):
            p = paths[i]           # order of nodes
            ordered = self.xy[p]    # ordered nodes
            # find cost of that order by the sum of euclidean distances between points (i) and (i+1)
            cost = (((ordered[:-1] - ordered[1:])**2).sum(1)).sum()
            if cost < mindist:
                mindist = cost
                minidx = i
        opt_order = paths[minidx]
        new_x = self.xy[:, 0][opt_order]
        new_y = self.xy[:, 1][opt_order]
        # return
        tmp_pts.xy = np.array(list(zip(new_x, new_y)))
        return tmp_pts

    def remove_nans(self, inplace=False):
        """
        Remove the points containing nans values.
        """
        if inplace:
            tmp_pts = self
        else:
            tmp_pts = self.copy()
        # get indices to remove
        inds = np.logical_or(np.isnan(tmp_pts.xy[:, 0]),
                             np.isnan(tmp_pts.xy[:, 1]))
        inds = np.logical_or(inds, np.isnan(tmp_pts.v))
        # remove
        for ind in np.where(inds)[0][::-1]:
            tmp_pts.remove(ind)
        # return
        if not inplace:
            return tmp_pts

    def augment_resolution(self, fact=2, interp='linear', inplace=False):
        """
        Augment the temporal resolution of the points.
        Only have sense if points are sorted to set some kind of trajectory.

        Parameters
        ----------
        fact : integer
            Resolution augmentation needed (default is '2', for a result
            profile with twice more points)
        interp : string in ['linear', 'nearest', slinear', 'quadratic, 'cubic']
            Specifies the kind of interpolation as a string
            (Default is 'linear'). slinear', 'quadratic' and 'cubic' refer
            to a spline interpolation of first, second or third order.
        inplace bool
            .

        Note
        ----
        If masked values are present, they are interpolated as well, using the
        surrounding values.
        """
        if inplace:
            tmp_pts = self
        else:
            tmp_pts = self.copy()
        is_v = len(self.v) == len(self.xy)
        # get and interpolate
        from .profile import Profile
        tmp_x = Profile(list(range(len(self.xy))), self.xy[:, 0])
        tmp_x.augment_resolution(fact=fact, interp=interp, inplace=True)
        tmp_y = Profile(list(range(len(self.xy))), self.xy[:, 1])
        tmp_y.augment_resolution(fact=fact, interp=interp, inplace=True)
        if is_v:
            tmp_v = Profile(list(range(len(self.xy))), self.v)
            tmp_v.augment_resolution(fact=fact, interp=interp, inplace=True)
        # store
        tmp_pts.xy = list(zip(tmp_x.y, tmp_y.y))
        if is_v:
            tmp_pts.v = tmp_v.y
        if not inplace:
            return tmp_pts

    def smooth(self, tos='gaussian', size=None, inplace=False, **kw):
        """
        Return a smoothed points field.

        Parameters
        ----------
        tos : string, optional
            Type of smoothing, can be 'uniform' (default), 'gaussian'
            or 'lowpass'.
        size : number, optional
            radius of the smoothing for 'uniform',
            radius of the smoothing for 'gaussian',
            cut off frequency for 'lowpass'
            Default are 3 for 'uniform',  1 for 'gaussian' and 0.1 for '
            lowpass'.
        inplace : boolean
            If 'False', return a smoothed points field
            else, smooth in place.
        kw : dic
            Additional parameters for ndimage methods
            (See ndimage documentation)
        """
        if not isinstance(tos, STRINGTYPES):
            raise TypeError("'tos' must be a string")
        if size is None and tos == 'uniform':
            size = 3
        elif size is None and tos == 'gaussian':
            size = 1
        elif size is None and tos == 'lowpass':
            size = 0.1
        # default smoothing border mode to 'nearest'
        if tos in ['uniform', 'gaussian'] and 'mode' not in list(kw.keys()):
            kw.update({'mode': 'nearest'})
        # getting data
        if not inplace:
            tmp_pts = self.copy()
        y = self.xy[:, 1]
        x = self.xy[:, 0]
        if len(self.v) != 0:
            v = self.v
        # smoothing
        if tos == "uniform":
            x = ndimage.uniform_filter(x, size, **kw)
            y = ndimage.uniform_filter(y, size, **kw)
            if len(self.v) != 0:
                v = ndimage.uniform_filter(v, size, **kw)
        elif tos == "gaussian":
            x = ndimage.gaussian_filter(x, size, **kw)
            y = ndimage.gaussian_filter(y, size, **kw)
            if len(self.v) != 0:
                v = ndimage.gaussian_filter(v, size, **kw)
        elif tos == 'lowpass':
            from scipy import signal
            x = self.xy[:, 0]
            y = self.xy[:, 1]
            N = 2
            Wn = size
            B, A = signal.butter(N, Wn, output='ba')
            x = signal.filtfilt(B, A, x)
            y = signal.filtfilt(B, A, y)
            if len(self.v) != 0:
                v = signal.filtfilt(B, A, v)

        else:
            raise ValueError("'tos' must be 'uniform', 'gaussian' or "
                             "'lowpass'")
        # storing
        if inplace:
            self.xy[:, 0] = x
            self.xy[:, 1] = y
            if len(self.v) != 0:
                self.v = v
        else:
            tmp_pts.xy[:, 0] = x
            tmp_pts.xy[:, 1] = y
            if len(tmp_pts.v) != 0:
                tmp_pts.v = v
            return tmp_pts

    def _display(self, kind=None, axe_x=None, axe_y=None, axe_color=None,
                 **plotargs):
        if len(self.xy) == 0:
            return None
        # x values
        if axe_x == 'x' or axe_x is None:
            x_values = self.xy[:, 0]
        elif axe_x == 'y':
            x_values = self.xy[:, 1]
        elif axe_x == 'v':
            if len(self.v) != len(self.xy):
                raise Exception()
            x_values = self.v
        else:
            raise ValueError()
        # y values
        if axe_y == 'x':
            y_values = self.xy[:, 0]
        elif axe_y == 'y' or axe_y is None:
            y_values = self.xy[:, 1]
        elif axe_y == 'v':
            if len(self.v) != len(self.xy):
                raise Exception()
            y_values = self.v
        else:
            raise ValueError()
        # color values
        if axe_color == 'x':
            color_values = self.xy[:, 0]
        elif axe_color == 'y':
            color_values = self.xy[:, 1]
        elif axe_color in [None, 'v']:
            if len(self.v) != len(self.xy):
                color_values = None
                if ('c' not in plotargs.keys()
                    and 'color' not in plotargs.keys()):
                    plotargs['color'] = 'k'
            color_values = self.v
        else:
            raise ValueError()
        # check c and color
        if 'c' in plotargs.keys() and 'color' in plotargs.keys():
            plotargs.pop('c')
        # display
        dp = pplt.Displayer(x=x_values, y=y_values, values=color_values,
                            kind=kind, **plotargs)
        ax = dp.draw()
        return ax

    def display(self, kind=None, axe_x=None, axe_y=None, axe_color=None,
                **plotargs):
        """
        Display the set of points.

        Parameters
        ----------
        kind : string, optional
            Can be 'plot' (default if points have not values).
            or 'scatter' (default if points have values).
            or 'colored_plot'.
        axe_x, axe_y, axe_color : strings in ['x', 'y', 'v']
            To determine wich value has to be plotted along which axis, and
            whith value is used to color the scattered points.
            Default plot 'y' to 'x' with colors from 'v'.
        **plotargs : dic
            Additionnal arguments sent to 'plot' or 'scatter'
        """
        # default values
        if axe_x is None:
            if axe_y != 'x':
                axe_x = 'x'
            else:
                axe_x = 'y'
        if axe_y is None:
            if axe_x != 'y':
                axe_y = 'y'
            else:
                axe_y = 'x'
        if axe_color is None:
            axes = ['x', 'y', 'v']
            try:
                axes.remove(axe_x)
                axes.remove(axe_y)
            except ValueError:
                axes = ['v']
            axe_color = axes[0]

        # display the values
        plot = self._display(kind, axe_x=axe_x, axe_y=axe_y,
                             axe_color=axe_color, **plotargs)
        if len(self.v) != 0 and kind is not 'plot':
            cb = plt.colorbar(plot)
            cb.set_label(self.unit_v.strUnit())
            # cb label
            if axe_color == 'x':
                cb.set_label('X ' + self.unit_x.strUnit())
            elif axe_color == 'y':
                cb.set_label('Y ' + self.unit_y.strUnit())
            else:
                cb.set_label('V ' + self.unit_v.strUnit())
        # x axis label
        if axe_x == 'x':
            plt.xlabel('X ' + self.unit_x.strUnit())
        elif axe_x == 'y':
            plt.xlabel('Y ' + self.unit_y.strUnit())
        else:
            plt.xlabel('V ' + self.unit_v.strUnit())
        # y axis label
        if axe_y == 'x':
            plt.ylabel('X ' + self.unit_x.strUnit())
        elif axe_y == 'y':
            plt.ylabel('Y ' + self.unit_y.strUnit())
        else:
            plt.ylabel('V ' + self.unit_v.strUnit())
        if self.name is None:
            plt.title('Set of points')
        else:
            plt.title(self.name)
        plt.axis('equal')
        return plot

    def display3D(self, kind='plot', xlabel='', ylabel='', zlabel='',
                  title='', **plotargs):
        """
        Display the points on a 3D graph.

        Parameters
        ----------
        kind : string, optional
            Kind of graph to use, can be 'plot' or 'surf'.
        xlabel, ylabel, zlabel : string, optional
            Label fo each axis (respectively 'x', 'y', and 'v')
        title : strin, optional
            Title
        **plotargs :
            Additional parameters feeded to matplotlib
        """
        # create 3D plot
        ax = plt.gca(projection='3d')
        # display data
        if kind == 'plot':
            ax.plot(self.xy[:, 0], self.xy[:, 1], self.v, **plotargs)
        elif kind == 'surf':
            ax.plot_trisurf(self.xy[:, 0], self.xy[:, 1], self.v)
        else:
            raise ValueError()
        # labels
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_zlabel(zlabel)
        # title
        ax.set_title(title)
        #
        plt.tight_layout()
        return ax

    def export_to_profile(self, axe_x='x', axe_y='y'):
        """
        Export the unsorted point object to a sorted Profile object.

        Parameters
        ----------
        axe_x, axe_y : strings in ['x', 'y', 'v']
            Which value used to construct the profile
        """
        # check
        if axe_x not in ['x', 'y', 'v']:
            raise ValueError()
        if axe_y not in ['x', 'y', 'v']:
            raise ValueError()
        # get data
        if axe_x == 'x':
            x = self.xy[:, 0]
            unit_x = self.unit_x
        elif axe_x == 'y':
            x = self.xy[:, 1]
            unit_x = self.unit_y
        else:
            x = self.v
            unit_x = self.unit_v
        if axe_y == 'x':
            y = self.xy[:, 0]
            unit_y = self.unit_x
        elif axe_y == 'y':
            y = self.xy[:, 1]
            unit_y = self.unit_y
        else:
            y = self.v
            unit_y = self.unit_v
        # construct profile
        from .profile import Profile
        prof = Profile(x=x, y=y, mask=False, unit_x=unit_x, unit_y=unit_y)
        return prof



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .file_operation import export_to_file, export_to_matlab, export_to_vtk,\
    import_from_file, import_from_IM7, import_from_IM7s, import_from_VC7,\
    import_from_VC7s, check_path, \
    import_sf_from_ascii, import_vf_from_ascii, import_vfs_from_ascii, \
    IM7_to_imt, VC7_to_imt, imts_to_imt, import_pts_from_ascii,\
    import_from_picture, import_from_pictures, export_to_picture,\
    export_to_pictures, export_to_ascii, import_profile_from_ascii,\
    check_path, import_from_matlab, import_from_video, export_to_video



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import re
import unum
import unum.units as units
try:
    units.counts = unum.Unum.unit('counts')
    units.px = unum.Unum.unit('px')
except:
    pass
from ..utils.types import ARRAYTYPES, INTEGERTYPES, NUMBERTYPES, STRINGTYPES


def make_unit(string):
    """
    Function helping for the creation of units. For more details, see the
    Unum module documentation.

    Parameters
    ----------
    string : string
        String representing some units.

    Returns
    -------
    unit : unum.Unum object
        unum object representing the given units

    Examples
    --------
    >>> make_unit("m/s")
    1 [m/s]
    >>> make_unit("N/m/s**3")
    1 [kg/s4]
    """
    if string == "":
        return unum.Unum({})
    # Safe check
    forbidden = ['import', '=', '\n', ';', ':', '"', "'", "open"]
    forbidden += [f'\w{key}\w' for key in globals().keys()]
    for f in forbidden:
        if f in string:
            raise Exception("Unauthorized string: {}".format(f))
    if re.match("[a-zA-Z]+\(.*\)", string):
        raise Exception("No call allowed here")
    # Replace '^' by '**'
    string = string.replace('^', '**')
    # exec
    env = {}
    exec("from unum.units import *;res = {}".format(string), env)
    return env['res']


def make_unit_old(string):
    """
    Function helping for the creation of units. For more details, see the
    Unum module documentation.

    Parameters
    ----------
    string : string
        String representing some units.

    Returns
    -------
    unit : unum.Unum object
        unum object representing the given units

    Examples
    --------
    >>> make_unit("m/s")
    1 [m/s]
    >>> make_unit("N/m/s**3")
    1 [kg/s4]
    """
    if len(string) == 0:
        return unum.Unum({})
    brackets = ['(', ')', '[', ']']
    symbambig = {"**": "^"}
    operators = ['*', '^', '/']
    toremove = [' ']

    def spliting(string):
        """
        Split the given string to elemental brick.
        """
        # suppression symboles inutiles
        for symb in toremove:
            string = string.replace(symb, '')
        # remplacement symboles ambigues
        for key in symbambig:
            string = string.replace(key, symbambig[key])
        # découpage de la chaine de caractère
        pieces = [string]
        for symb in operators + brackets:
            j = len(pieces)-1
            while True:
                if (pieces[j].find(symb) != -1 and
                        len(pieces[j]) != len(symb)):
                    splitpiece = list(pieces[j].partition(symb))
                    pieces[j:j+1] = splitpiece
                    j += 2
                else:
                    j -= 1
                if j < 0:
                    break
        # Suppression des caractères nuls
        for j in np.arange(len(pieces)-1, -1, -1):
            if pieces[j] == '':
                pieces[j:j+1] = []
        # Add power where needed
        for i, piece in enumerate(pieces):
            match = re.match('([a-zA-Z]+)([0-9]+)', piece)
            if match is not None:
                pieces[i:i+1] = [match.groups()[0], "^", match.groups()[1]]
        return pieces

    def app_brackets(strlist):
        """
        Apply all the brackets.
        """
        level = 0
        levels = []
        for i in np.arange(len(strlist)):
            if strlist[i] in ['(', '[']:
                level += 1
            elif strlist[i-1] in [')', ']'] and i-1 > 0:
                level -= 1
            if level < 0:
                raise ValueError("I think you have a problem"
                                 " with your brackets, dude")
            levels.append(level)
        # boucle sur les parenthèses
        while True:
            if len(strlist) == 1:
                return app_one_bracket(strlist)
            first_ind = levels.index(np.max(levels))
            last_ind = None
            for i in np.arange(first_ind, len(levels)):
                if levels[i] < levels[first_ind]:
                    last_ind = i-1
                    break
            if last_ind is None:
                last_ind = len(levels) - 1
            if np.all(np.array(levels) == 0):
                result = app_one_bracket(strlist)
                return result
            else:
                strlist[first_ind:last_ind+1] \
                    = [app_one_bracket(strlist[first_ind+1:last_ind])]
            levels[first_ind:last_ind+1] = [np.max(levels)-1]

    def app_one_bracket(stringlist):
        """
        Apply one bracket.
        """
        stringlist = list(stringlist)
        # traitement des nombres
        for j in np.arange(len(stringlist)-1, -1, -1):
            try:
                stringlist[j] = float(stringlist[j])
                if stringlist[j] % 1 == 0:
                    stringlist[j] = int(stringlist[j])
            except:
                pass
        # traitement des unités
        for j in np.arange(len(stringlist)-1, -1, -1):
            if not isinstance(stringlist[j], (unum.Unum, NUMBERTYPES)):
                if not stringlist[j] in operators + brackets:
                    if not stringlist[j] in unum.Unum.getUnitTable():
                        raise ValueError("I don't know this unit : '{}',"
                                         "little bird. (or maybe it's an "
                                         "operator i'm missing...)."
                                         .format(stringlist[j]))
                    stringlist[j] = unum.Unum({stringlist[j]: 1})
        # traitement des opérateurs
        liste = stringlist
        # ^
        for ind in np.arange(len(liste)-1, 0, -1):
            if isinstance(liste[ind], unum.Unum):
                continue
            if liste[ind] == '^':
                liste[ind - 1:ind + 2] = [liste[ind - 1]**liste[ind + 1]]
                ind -= 2
                if ind < 0:
                    break
        # /
        for ind in np.arange(len(liste)-1, 0, -1):
            if isinstance(liste[ind], unum.Unum):
                continue
            if liste[ind] == '/':
                liste[0:1] = [liste[0]/liste[ind + 1]]
                liste[ind:ind+2] = []
                #liste[ind - 1:ind + 2] = [liste[ind - 1]/liste[ind + 1]]
                ind -= 2
                if ind < 0:
                    break
        # *
        for ind in np.arange(len(liste)-1, 0, -1):
            if isinstance(liste[ind], unum.Unum):
                continue
            if liste[ind] == '*':
                liste[ind - 1:ind + 2] = [liste[ind - 1]*liste[ind + 1]]
                ind -= 2
                if ind < 0:
                    break
        return liste[0]
    strlist = spliting(string)
    unity = app_brackets(strlist)
    return unity



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import warnings
import numpy as np
import shutil
import os
from os.path import join
import copy
import re
from ..utils.types import ARRAYTYPES
from ..utils.progresscounter import ProgressCounter

try:
    import colorama
except ImportError:
    colorama = None

class Files(object):
    def __init__(self):
        """
        Class representing a bunch of files (and/or folders)
        """
        self.paths = []
        self.exist = []
        self.isdir = []

    def __add__(self, obj):
        if isinstance(obj, Files):
            tmp_files = self.copy()
            tmp_files.paths += obj.paths
            tmp_files.exist += obj.exist
            tmp_files.isdir += obj.isdir
            return tmp_files

    def __repr__(self):
        text = self.get_tree_representation(max_file_list=10, hide_top=True)
        return text

    def get_tree_representation(self, max_file_list=10, hide_top=False):
        self.build_tree()
        # properties
        tab_color = ""
        folder_color = ""
        file_color = ""
        file_number_color = ""
        tab0 = "|"
        tab1 = tab0 + ">" + tab0
        tab2 = tab0 + " "
        separator = "/"
        heading_separator = "="
        if colorama is not None:
            tab_color = colorama.Fore.BLACK + colorama.Style.BRIGHT
            tab0 = tab_color + tab0 + colorama.Style.NORMAL
            tab1 = tab_color + tab1 + colorama.Style.NORMAL
            tab2 = tab_color + tab2 + colorama.Style.NORMAL
            folder_color = colorama.Fore.BLACK
            file_color = colorama.Fore.GREEN
            file_number_color = colorama.Fore.CYAN
        folders_end_of_line = "\n"
        files_end_of_line = "\n"
        max_file_list = max_file_list

        # recursion function
        def get_info_for_folder(fold, tab1, tab2):
            if not isinstance(fold, dict):
                raise TypeError()
            # if too much files to display, display number of files
            if 'files' in list(fold.keys()):
                if len(fold['files']) > max_file_list:
                    yield (tab1 + file_number_color +
                           '[{} Files]'.format(len(fold['files'])) +
                           files_end_of_line)
            # loop recursively on folder's folders
            for key in list(fold.keys()):
                if key == 'files':
                    pass
                else:
                    yield tab1 + folder_color + key + folders_end_of_line
                    for thing in get_info_for_folder(fold[key], tab1, tab2):
                        yield tab2 + thing
            # loop on folder's files
            if 'files' in list(fold.keys()):
                if len(fold['files']) <= max_file_list:
                    for f in fold['files']:
                        yield tab1 + file_color + f + files_end_of_line
        # get
        text = ""
        tree = self.tree
        # add heading wih common  path
        if hide_top:
            curr_fold = self.tree
            while True:
                if isinstance(curr_fold, dict):
                    if len(list(curr_fold.keys())) == 1 and \
                            list(curr_fold.keys())[0] != "files":
                        text += list(curr_fold.keys())[0] + separator
                        curr_fold = curr_fold[list(curr_fold.keys())[0]]
                    else:
                        break
                else:
                    break
            # put a nice heading
            text = text[:-1]
            heading_sep = heading_separator*(len(text)+2)
            text = (tab0 + tab_color + heading_sep + tab0 +
                    folders_end_of_line +
                    tab0 + " " + text + " " + tab0 + folders_end_of_line +
                    tab_color + tab0 + tab_color + heading_sep + tab0 +
                    folders_end_of_line)
            # update folder
            tree = curr_fold
        # display the tree
        for thing in get_info_for_folder(tree, tab1=tab1,
                                         tab2=tab2):
            text += thing
        return text

    def copy(self):
        return copy.deepcopy(self)

    def add_file(self, path):
        # check argument type
        try:
            path = str(path)
        except:
            raise TypeError()
        # check if valid path or not
        path = os.path.normpath(path)
        if os.path.exists(path):
            self.paths.append(path)
            self.exist.append(True)
            if os.path.isdir(path):
                self.isdir.append(True)
            else:
                self.isdir.append(False)
        elif os.access(os.path.dirname(path), os.W_OK):
            self.paths.append(path)
            self.exist.append(False)
            self.isdir.append(None)
        else:
            raise Exception()

    def remove_files(self, arg):
        """
        Remove some files from the files set.

        Parameters
        ----------
        arg : integer, regex or array of integer or regex
            If integer, remove the associated path,
            if a regex, remove the paths that match,
            if an array, delete paths for each element.
        """
        if isinstance(arg, int):
            ind = arg
            del self.paths[ind]
            del self.exist[ind]
            del self.isdir[ind]
        elif isinstance(arg, ARRAYTYPES):
            for thing in arg:
                self.remove_files(thing)
        elif isinstance(arg, str):
            for i in np.arange(len(self.paths) - 1, -1, -1):
                res = re.match(arg, self.paths[i])
                if res is not None:
                    print("remove {}".format(self.paths[i]))
                    self.remove_files(i)
        else:
            raise TypeError()

    def remove_empty_directories():
        """
        """
        pass

    def load_files_from_regex(self, rootpath, regex, load_files=True,
                              load_dirs=True, depth='all'):
        """
        Load existing files from a regular expression.
        """
        # get folder names
        paths = []
        isdir = []
        exist = []
        rootpath = os.path.normpath(rootpath)
        for root, dirs, files in os.walk(rootpath, topdown=True):
            # check depth
            if depth != "all":
                tmp_depth = (root[len(rootpath) + len(os.path.sep):]
                             .count(os.path.sep))
                if depth < tmp_depth:
                    break
            # load direcories
            if load_dirs:
                for d in dirs:
                    fullpath = join(root, d)
                    if re.match(regex, fullpath):
                        paths.append(fullpath)
                        isdir.append(True)
                        exist.append(True)
            # load files
            if load_files:
                for f in files:
                    fullpath = join(root, f)
                    if re.match(regex, fullpath):
                        paths.append(fullpath)
                        isdir.append(False)
                        exist.append(True)
        # store
        self.paths += paths
        self.isdir += isdir
        self.exist += exist

    def build_tree(self):
        """
        Build a file tree
        """
        dic = {}
        for isdir, path in zip(self.isdir, self.paths):
            sep_path = path.split(os.path.sep)
            tmp_dic = dic
            sep_path_len = len(sep_path)
            for i, fold in enumerate(sep_path):
                # folders
                if isdir or i != sep_path_len - 1:
                    if fold in list(tmp_dic.keys()):
                        tmp_dic = tmp_dic[fold]
                    else:
                        new_dic = {}
                        tmp_dic[fold] = new_dic
                        tmp_dic = new_dic
                # files ()
                else:
                    if 'files' in list(tmp_dic.keys()):
                        tmp_dic['files'].append(fold)
                    else:
                        tmp_dic['files'] = [fold]
        # store
        self.tree = dic

    def delete_existing_files(self, recursive=False):
        tmp_isdir = list(np.array(self.isdir)[np.array(self.exist)])
        tmp_paths = list(np.array(self.paths)[np.array(self.exist)])
        nmb_dir = np.sum(tmp_isdir)
        nmb_files = len(tmp_paths) - nmb_dir
        print("+++ Ready to remove {} files and {} directories "
              .format(nmb_files, nmb_dir))
        while True:
            rep = input("+++ Okay with that ? ('o', 'n') \n+++ ")
            if rep in ['o', 'O', 'y', 'Y', 'oui', 'Oui',
                       'Yes', 'yes', 'YES', 'OUI']:
                rep = True
                break
            elif rep in ['n', 'N', 'No', 'no', 'non',
                         'Non', 'NON', 'NO']:
                rep = False
                break
        # remove if necessary
        if rep:
            print("")
            PG = ProgressCounter(init_mess="Begin cleaning",
                                 nmb_max=nmb_files,
                                 name_things='files',
                                 perc_interv=10)
            # remove files
            for i, p in enumerate(np.array(tmp_paths)):
                if tmp_isdir[i]:
                    continue
                PG.print_progress()
                os.remove(p)

            # remove dirs
            for i, p in enumerate(np.array(tmp_paths)):
                if tmp_isdir[i]:
                    # force recursive removing
                    if recursive:
                        shutil.rmtree(p)
                    else:
                        # else, check if each folder is empty
                        try:
                            os.rmdir(p)
                        except WindowsError:
                            print("+++ Following folder is not empty\n"
                                  "{}".format(p))
                            while True:
                                rep = input("+++ Delete anyway ? ('o', 'n') "
                                            "\n+++ ")
                                if rep in ['o', 'O', 'y', 'Y', 'oui', 'Oui',
                                           'Yes', 'yes', 'YES', 'OUI']:
                                    rep = True
                                    break
                                elif rep in ['n', 'N', 'No', 'no', 'non',
                                             'Non', 'NON', 'NO']:
                                    rep = False
                                    break
                            if rep is True:
                                shutil.rmtree(p)
            #
            for i in range(len(self.exist)):
                self.exist[i] = False


def remove_files_in_dirs(rootpath, dir_regex, file_regex,
                         depth='all', remove_dir=False, remove_files=True):
    """
    make a recursive search for directories satisfying "dir_regex'
    from "rootpath", and remove all the files satisfying 'file_regex' in it.

    Parameters
    ----------
    rootpath : string
        Path where to begin searching.
    dir_regex : string
        Regular expression matching the directory where we want to remove
        stuff.
    file_regex : string
        Regular expression matching the files we want to delete.
    depth : integer or 'all'
        Number of directory layer to go through.
    remove_dir :



    """
    warnings.warn("Deprecated, use 'Files' class instead")
    # ### TODO : add the possibility to remove empty directories
    # get dirs
    dir_paths = []
    nmb_files = []
    file_paths = []
    nmb_tot_files = 0
    for root, dirs, files in os.walk(rootpath):
        nmb_tot_files += len(files)
        if re.match(dir_regex, root):
            tmp_nmb_files = 0
            for f in files:
                if re.match(file_regex, f):
                    tmp_nmb_files += 1
                    file_paths.append(os.path.join(root, f))
            # check if there is actuelly files in there
            if not tmp_nmb_files == 0:
                nmb_files.append(tmp_nmb_files)
                dir_paths.append(root)
    # ask before deletion
    print("")
    print("+++ Checked {} files".format(nmb_tot_files))
    if np.sum(nmb_files) == 0:
        print("+++ Nothing to delete")
        return None
    print("+++ Ready to remove {} files in directories :"
          .format(np.sum(nmb_files)))
    for i in range(len(dir_paths)):
        print("+++    [{} files] {}".format(nmb_files[i], dir_paths[i]))
    while True:
        rep = input("+++ Okay with that ? ('o', 'n') \n+++ ")
        if rep in ['o', 'O', 'y', 'Y', 'oui', 'Oui',
                   'Yes', 'yes', 'YES', 'OUI']:
            rep = True
            break
        elif rep in ['n', 'N', 'No', 'no', 'non', 'Non', 'NON', 'NO']:
            rep = False
            break
    # remove if necessary
    if rep:
        print("")
        PG = ProgressCounter(init_mess="Begin cleaning",
                             nmb_max=len(file_paths),
                             name_things='files',
                             perc_interv=10)
        for p in file_paths:
            PG.print_progress()
            os.remove(p)



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2013-2018 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# This submodule is inspired from https://github.com/galaunay/python_video_stab

from ..core.profile import Profile
from ..core.temporalscalarfields import TemporalScalarFields
from ..core.scalarfield import ScalarField
from ..utils.progresscounter import ProgressCounter
from ..plotlib import get_color_cycles
import numpy as np
import matplotlib.pyplot as plt
import cv2


class Stabilizer(object):
    """
    Stabilizer for TemporalFields, based on ORB implementation of opencv.
    """

    def __init__(self, obj, orb_kwargs={}, mode='continuous'):
        """
        Stabilizer for TemporalFields, based on ORB implementation of opencv.

        Parameters
        ==========
        obj: TemporalFields object
            Fields on wich performing the stabilization.
        orb_kwargs: dic
            Additional arguments for the ORB feature detectior
            (See opencv documentation for more information)
        mode: string in ['continuous', 'from_first']
            If 'continuous', stabilize each frame in regards of the previous
            one.
            If 'from_first', stabilize all frame from the first one.
        """
        self.obj = obj
        self.kp_kwargs = orb_kwargs
        self.kp_detector = cv2.ORB_create(**orb_kwargs)
        self.kp_keypoints = []
        self.mode = mode
        self.raw_transform = None
        self.smoothed_transform = None
        self.raw_trajectory = None
        self.smoothed_trajectory = None
        self.stabilized_obj = None

    def _compute_transform(self, dense=False, opt_flow_args={}, verbose=False):
        """
        Use the feature points to compute the deformation transformation for
        each frames.

        Parameters
        ==========
        dense: boolean
            If True, use dens optical flow (track every pixel instead
            of tracking features). This method give better results when the
            frames don't show well-defined features, but is slower.
        opt_flow_args: dic
            Additional arguments for the optical flow functions
            ('calcOpticalFLowPyrLK' or 'calcOpticalFlowFarneback',
            see opencv documentation.
        """
        # verbose
        if verbose:
            pg = ProgressCounter(init_mess="Computing transforms",
                                 nmb_max=len(self.obj) - 1,
                                 name_things="transforms")
        prev_to_cur_transform = []
        prev_im = self.obj[0].values
        for i in np.arange(1, len(self.obj)):
            cur_im = self.obj[i].values
            if dense:
                raise Exception('Not working for drops...')
                args = {'pyr_scale': 0.5,
                        'levels': 3,
                        'winsize': 15,
                        'iterations': 3,
                        'poly_n': 5,
                        'poly_sigma': 1.2,
                        'flags': 0}
                args.update(opt_flow_args)
                flow = cv2.calcOpticalFlowFarneback(
                    prev_im,
                    cur_im,
                    None, **args)
                # get mean displacement from dxs, dys
                dxs = flow[..., 0]
                dys = flow[..., 1]
                dx = np.median(dxs)
                dy = np.median(dys)
                da = 0
            else:
                # detect keypoints
                prev_kps = self.kp_detector.detect(prev_im)
                self.kp_keypoints.append(prev_kps)
                if len(prev_kps) == 0:
                    raise Exception(f'No features found on the {i}th frame, '
                                    'you should try reducing the '
                                    '\'fastThreshold\' value.')
                prev_kps = np.array(
                    [kp.pt for kp in prev_kps],
                    dtype='float32').reshape(-1, 1, 2)
                # calc flow of movement
                cur_kps, status, err = cv2.calcOpticalFlowPyrLK(
                    prev_im,
                    cur_im,
                    prev_kps,
                    None,
                    **opt_flow_args)
                # storage for keypoints with status 1
                prev_matched_kp = []
                cur_matched_kp = []
                for i, matched in enumerate(status):
                    # store coords of keypoints that appear in both
                    if matched:
                        prev_matched_kp.append(prev_kps[i])
                        cur_matched_kp.append(cur_kps[i])
                # estimate partial transform
                transform = cv2.estimateRigidTransform(
                    np.array(prev_matched_kp),
                    np.array(cur_matched_kp),
                    False)
                if transform is not None:
                    dx = transform[0, 2]
                    dy = transform[1, 2]
                    da = np.arctan2(transform[1, 0], transform[0, 0])
                else:
                    dx = dy = da = 0

            # store transform
            prev_to_cur_transform.append([dx, dy, da])
            if self.mode == 'continuous':
                # set current frame to prev frame for use in next iteration
                prev_im = cur_im
            # verbose
            if verbose:
                pg.print_progress()
        # store resulting traj and transform
        if self.mode == 'continuous':
            self.raw_transform = np.array(prev_to_cur_transform)
            self.raw_trajectory = np.cumsum(prev_to_cur_transform, axis=0)
        elif self.mode == 'from_first':
            prev_to_cur_transform = np.array(prev_to_cur_transform)
            self.raw_transform = np.concatenate(([prev_to_cur_transform[0]],
                                                 prev_to_cur_transform[1::] -
                                                 prev_to_cur_transform[0:-1]))
            self.raw_trajectory = np.array(prev_to_cur_transform)
        else:
            raise ValueError()

    def _smooth_transform(self, smooth_size=30, verbose=False):
        """
        Smooth the computed transformations to get smoother transitions in the
        resulting video.

        Parameters
        ==========
        smooth_size: integer
            Size of the smoothing to use (default to 30)
        """
        # check
        if self.raw_transform is None:
            raise Exception('You should compute the transformation first.')
        # verbose
        if verbose:
            print("=== Smoothing transforms ===")
        self.smoothed_trajectory = self.raw_trajectory.copy()
        for i in range(3):
            tmp_prof = Profile(self.obj.times[:-1], self.raw_trajectory[:, i])
            tmp_prof.smooth(tos='gaussian', size=smooth_size,
                            inplace=True)
            self.smoothed_trajectory[:, i] = tmp_prof.y
        self.smoothed_transform = (self.raw_transform
                                   + (self.smoothed_trajectory
                                      - self.raw_trajectory))

    def _apply_transform(self, border_type='black', border_size=0,
                         verbose=False):
        """
        Apply the computed transformation to the fields.

        Parameters:
        ===========
        border_type: string in ['black', 'reflect', 'replicate']
            Type of borders.
        border_size: integer
            Size in pixels of the additional border to use
            (to avoid loosing part of the frames during the transformation).
        """
        # checks
        if self.smoothed_transform is None:
            raise Exception('You should compute and smooth the transformation'
                            ' first.')
        # verbose
        if verbose:
            pg = ProgressCounter(init_mess="Applying transform",
                                 nmb_max=len(self.obj),
                                 name_things="transforms")
        # checks
        border_modes = {
            'black': cv2.BORDER_CONSTANT,
            'reflect': cv2.BORDER_REFLECT,
            'replicate': cv2.BORDER_REPLICATE
        }
        border_mode = border_modes[border_type]
        # get im shape
        h, w = self.obj.shape
        h += 2 * border_size
        w += 2 * border_size
        # create result holder
        res_tsf = TemporalScalarFields()
        dx = self.obj.dx
        x0, xf = self.obj.axe_x[0], self.obj.axe_x[-1]
        dy = self.obj.dy
        y0, yf = self.obj.axe_y[0], self.obj.axe_y[-1]
        axe_x = np.arange(x0 - border_size*dx, xf + border_size*dx + dx, dx)
        axe_y = np.arange(y0 - border_size*dy, yf + border_size*dy + dx, dy)
        # main loop
        transf = np.concatenate(([[0, 0, 0]], self.smoothed_transform))
        for i in range(len(self.obj)):
            # build transformation matrix
            transform = np.zeros((2, 3))
            transform[0, 0] = np.cos(transf[i][2])
            transform[0, 1] = -np.sin(transf[i][2])
            transform[1, 0] = np.sin(transf[i][2])
            transform[1, 1] = np.cos(transf[i][2])
            transform[0, 2] = transf[i][0]
            transform[1, 2] = transf[i][1]
            # apply transform
            bordered_im = cv2.copyMakeBorder(
                self.obj[i].values,
                top=border_size * 2,
                bottom=border_size * 2,
                left=border_size * 2,
                right=border_size * 2,
                borderType=border_mode,
                value=[0, 0, 0])
            transformed_im = cv2.warpAffine(
                bordered_im,
                transform,
                (w + border_size * 2, h + border_size * 2),
                borderMode=border_mode)
            transformed_im = transformed_im[border_size:(transformed_im.shape[0]
                                                         - border_size),
                                            border_size:(transformed_im.shape[1]
                                                         - border_size)]
            # store
            im = ScalarField()
            im.import_from_arrays(axe_x=axe_x,
                                  axe_y=axe_y,
                                  values=transformed_im,
                                  unit_x=self.obj.unit_x,
                                  unit_y=self.obj.unit_y,
                                  unit_values=self.obj.unit_values,
                                  dtype=np.uint8)
            res_tsf.add_field(im, time=self.obj.times[i],
                              unit_times=self.obj.unit_times,
                              copy=False)
            # verbose
            if verbose:
                pg.print_progress()
        self.stabilized_obj = res_tsf

    def _apply_transform_to_point(self, pts, from_frame=0, verbose=True):
        """
        Apply the transformation to pts of the first image.

        Parameters
        ==========
        pts: Nx2 array
            Point of the first array
        from_frame: integer
            Frame from wich the tracking should take its source.
        """
        # checks
        if self.smoothed_transform is None:
            raise Exception('You should compute and smooth the transformation'
                            ' first.')
        transform = self.raw_transform
        pts = np.array(pts)
        if pts.ndim == 1:
            pts = np.array([pts])
        if from_frame != 0:
            raise Exception('Not implemented yet')
        # verbose
        if verbose:
            pg = ProgressCounter(init_mess="Tracking points motion",
                                 nmb_max=len(self.obj) - 1,
                                 name_things="transforms")
            n = 0
        all_new_pts = []
        for pt in pts:
            new_pts = [np.array(pt)]
            for tr in transform:
                tr_mat = np.zeros((3, 3))
                tr_mat[0, 0] = np.cos(tr[2])
                tr_mat[0, 1] = -np.sin(tr[2])
                tr_mat[0, 2] = 0
                tr_mat[1, 0] = np.sin(tr[2])
                tr_mat[1, 1] = np.cos(tr[2])
                tr_mat[1, 2] = 0
                tr_mat[0, 2] = tr[0]
                tr_mat[1, 2] = tr[1]
                tr_mat[2, 2] = 1
                pt = np.array([new_pts[-1][1], new_pts[-1][0]],
                              dtype=np.float32)
                pt = np.reshape(pt, (1, 1, 2))
                tmp_pts = cv2.perspectiveTransform(pt, tr_mat)
                tmp_pts = tmp_pts[0][0][0:2]
                tmp_pts = tmp_pts[::-1]
                new_pts.append(tmp_pts)
                # verbose
                if verbose:
                    n += 1
                    if n%len(pts) == 0:
                        pg.print_progress()
            all_new_pts.append(new_pts)
        # return
        return np.array(all_new_pts)

    def get_stabilized_obj(self, dense=False, smooth_size=30,
                           border_type='black', border_size=0,
                           verbose=False):
        """
        Stabilize the fields.

        Parameters
        ==========
        dense: boolean
            If True, use dens optical flow (track every pixel instead
            of tracking features).
        smooth_size: integer
            Size of the smoothing to use (default to 30)
        border_type: string in ['black', 'reflect', 'replicate']
            Type of borders.
        border_size: integer
            Size in pixels of the additional border to use
            (to avoid loosing part of the frames during the transformation).
        """
        self._compute_transform(dense=dense, verbose=verbose)
        self._smooth_transform(smooth_size=smooth_size, verbose=verbose)
        self._apply_transform(border_type=border_type, border_size=border_size,
                              verbose=verbose)
        return self.stabilized_obj

    def display_transform(self):
        """
        """
        # colors
        colors = get_color_cycles()
        # data
        data_dic = {'Raw transform': self.raw_transform,
                    'Raw trajectory': self.raw_trajectory,
                    'Smoothed trajectory': self.smoothed_trajectory,
                    'Smoothed transform': self.smoothed_transform}
        # plot
        for name, data in data_dic.items():
            fig, ax = plt.subplots(2, 1, sharex=True)
            plt.sca(ax[0])
            plt.plot(data[:, 0], label=f"{name} x",
                     color=colors[0])
            plt.plot(data[:, 1], label=f"{name} y",
                     color=colors[1])
            plt.legend()
            plt.sca(ax[1])
            plt.plot(data[:, 2], label=f"{name} $\\theta$",
                     color=colors[2])
            plt.legend()



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import unum
import copy
from ..utils.types import ARRAYTYPES, NUMBERTYPES, STRINGTYPES
from ..utils import make_unit


class Field(object):

    def __init__(self):
        self.__axe_x = np.array([], dtype=float)
        self.__axe_y = np.array([], dtype=float)
        self.__unit_x = make_unit('')
        self.__unit_y = make_unit('')
        self.xy_scale = make_unit("")

    def __iter__(self):
        for i, x in enumerate(self.axe_x):
            for j, y in enumerate(self.axe_y):
                yield [i, j], [x, y]

    @property
    def axe_x(self):
        return self.__axe_x

    @axe_x.setter
    def axe_x(self, new_axe_x):
        if not isinstance(new_axe_x, ARRAYTYPES):
            raise TypeError()
        new_axe_x = np.array(new_axe_x, dtype=float)
        # if new_axe_x.shape == self.__axe_x.shape or len(self.__axe_x) == 0:
        self.__axe_x = new_axe_x
        # else:
        #     raise ValueError()

    @axe_x.deleter
    def axe_x(self):
        raise Exception("Nope, can't do that")

    @property
    def dx(self):
        return self.axe_x[1] - self.axe_x[0]

    @property
    def axe_y(self):
        return self.__axe_y

    @axe_y.setter
    def axe_y(self, new_axe_y):
        if not isinstance(new_axe_y, ARRAYTYPES):
            raise TypeError()
        new_axe_y = np.array(new_axe_y, dtype=float)
        # if new_axe_y.shape == self.__axe_y.shape or len(self.__axe_y) == 0:
        self.__axe_y = new_axe_y
        # else:
        #     raise ValueError()

    @axe_y.deleter
    def axe_y(self):
        raise Exception("Nope, can't do that")

    @property
    def dy(self):
        return self.axe_y[1] - self.axe_y[0]

    @property
    def unit_x(self):
        return self.__unit_x

    @unit_x.setter
    def unit_x(self, new_unit_x):
        if isinstance(new_unit_x, unum.Unum):
            if np.isclose(new_unit_x.asNumber(), 1):
                self.__unit_x = new_unit_x
            else:
                raise ValueError()
        elif isinstance(new_unit_x, STRINGTYPES):
            self.__unit_x = make_unit(new_unit_x)
        else:
            raise TypeError()

    @unit_x.deleter
    def unit_x(self):
        raise Exception("Nope, can't do that")

    @property
    def unit_y(self):
        return self.__unit_y

    @unit_y.setter
    def unit_y(self, new_unit_y):
        if isinstance(new_unit_y, unum.Unum):
            if np.isclose(new_unit_y.asNumber(), 1):
                self.__unit_y = new_unit_y
            else:
                raise ValueError()
        elif isinstance(new_unit_y, STRINGTYPES):
            self.__unit_y = make_unit(new_unit_y)
        else:
            raise TypeError()

    @unit_y.deleter
    def unit_y(self):
        raise Exception("Nope, can't do that")

    @property
    def shape(self):
        return self.__axe_x.shape[0], self.__axe_y.shape[0]

    def check_consistency(self):
        """
        Raise errors if the field show some
        incoherences.
        """
        if not isinstance(self.axe_x, np.ndarray):
            raise Exception()
        if not isinstance(self.axe_y, np.ndarray):
            raise Exception()
        if not isinstance(self.unit_x, unum.Unum):
            raise Exception()
        if not isinstance(self.unit_y, unum.Unum):
            raise Exception()
        if not isinstance(self.xy_scale, unum.Unum):
            raise Exception()

    def copy(self):
        """
        Return a copy of the Field object.
        """
        return copy.deepcopy(self)

    def get_indice_on_axe(self, direction, value, kind='bounds'):
        """
        Return, on the given axe, the indices representing the positions
        surrounding 'value'.
        if 'value' is exactly an axe position, return just one indice.

        Parameters
        ----------
        direction : int
            1 or 2, for axes choice.
        value : number
        kind : string
            If 'bounds' (default), return the bounding indices.
            if 'nearest', return the nearest indice
            if 'decimal', return a decimal indice (interpolated)

        Returns
        -------
        interval : 2x1 or 1x1 array of integer
        """
        if not isinstance(direction, NUMBERTYPES):
            raise TypeError("'direction' must be a number.")
        if not (direction == 1 or direction == 2):
            raise ValueError("'direction' must be 1 or 2.")
        if not isinstance(value, NUMBERTYPES):
            raise TypeError("'value' must be a number.")
        if direction == 1:
            axe = self.axe_x
            if value < axe[0] or value > axe[-1]:
                raise ValueError("'value' is out of bound.")
        else:
            axe = self.axe_y
            if value < axe[0] or value > axe[-1]:
                raise ValueError("'value' is out of bound.")
        if not isinstance(kind, STRINGTYPES):
            raise TypeError()
        if kind not in ['bounds', 'nearest', 'decimal']:
            raise ValueError()
        # getting the borning indices
        ind = np.searchsorted(axe, value)
        if axe[ind] == value:
            inds = [ind, ind]
        else:
            inds = [int(ind - 1), int(ind)]
        # returning bounds
        if kind == 'bounds':
            return inds
        # returning nearest
        elif kind == 'nearest':
            if inds[0] == inds[1]:
                return inds[0]
            if np.abs(axe[inds[0]] - value) < np.abs(axe[inds[1]] - value):
                ind = inds[0]
            else:
                ind = inds[1]
            return int(ind)
        # returning decimal
        elif kind == 'decimal':
            if inds[0] == inds[1]:
                return inds[0]
            value_1 = axe[inds[0]]
            value_2 = axe[inds[1]]
            delta = np.abs(value_2 - value_1)
            return (inds[0]*np.abs(value - value_2)/delta +
                    inds[1]*np.abs(value - value_1)/delta)

    def get_points_around(self, center, radius, ind=False):
        """
        Return the list of points or the scalar field that are in a circle
        centered on 'center' and of radius 'radius'.

        Parameters
        ----------
        center : array
            Coordonate of the center point (in axes units).
        radius : float
            radius of the cercle (in axes units).
        ind : boolean, optional
            If 'True', radius and center represent indices on the field.
            if 'False', radius and center are expressed in axis unities.

        Returns
        -------
        indices : array
            Array contening the indices of the contened points.
            [(ind1x, ind1y), (ind2x, ind2y), ...].
            You can easily put them in the axes to obtain points coordinates
        """
        # checking parameters
        if not isinstance(center, ARRAYTYPES):
            raise TypeError("'center' must be an array")
        center = np.array(center, dtype=float)
        if not center.shape == (2,):
            raise ValueError("'center' must be a 2x1 array")
        if not isinstance(radius, NUMBERTYPES):
            raise TypeError("'radius' must be a number")
        if not radius > 0:
            raise ValueError("'radius' must be positive")
        # getting indice data when 'ind=False'
        if not ind:
            dx = self.axe_x[1] - self.axe_x[0]
            dy = self.axe_y[1] - self.axe_y[0]
            delta = (dx + dy)/2.
            radius = radius/delta
            center_x = self.get_indice_on_axe(1, center[0], kind='decimal')
            center_y = self.get_indice_on_axe(2, center[1], kind='decimal')
            center = np.array([center_x, center_y])
        # pre-computing somme properties
        radius2 = radius**2
        radius_int = radius/np.sqrt(2)
        # isolating possibles indices
        inds_x = np.arange(np.int(np.ceil(center[0] - radius)),
                           np.int(np.floor(center[0] + radius)) + 1)
        inds_y = np.arange(np.int(np.ceil(center[1] - radius)),
                           np.int(np.floor(center[1] + radius)) + 1)
        inds_x, inds_y = np.meshgrid(inds_x, inds_y)
        inds_x = inds_x.flatten()
        inds_y = inds_y.flatten()
        # loop on possibles points
        inds = []
        for i in np.arange(len(inds_x)):
            x = inds_x[i]
            y = inds_y[i]
            # test if the point is in the square 'compris' in the cercle
            if x <= center[0] + radius_int \
                    and x >= center[0] - radius_int \
                    and y <= center[1] + radius_int \
                    and y >= center[1] - radius_int:
                inds.append([x, y])
            # test if the point is the center
            elif all([x, y] == center):
                pass
            # test if the point is in the circle
            elif ((x - center[0])**2 + (y - center[1])**2 <= radius2):
                inds.append([x, y])
        return np.array(inds, subok=True)

    def scale(self, scalex=None, scaley=None, inplace=False,
              output_reverse=False):
        """
        Scale the Field.

        Parameters
        ----------
        scalex, scaley : numbers or Unum objects
            Scale for the axis
        inplace : boolean
            .
        """
        if inplace:
            tmp_f = self
        else:
            tmp_f = self.copy()
        # set xy_scale
        if scalex is not None and scaley is not None:
            tmp_f.xy_scale *= scalex/scaley
        elif scalex is not None:
            tmp_f.xy_scale *= scalex
        elif scaley is not None:
            tmp_f.xy_scale *= 1/scaley
        # x
        reversex = False
        if scalex is None:
            pass
        elif isinstance(scalex, NUMBERTYPES):
            tmp_f.axe_x *= scalex
            if scalex < 0:
                reversex = True
        elif isinstance(scalex, unum.Unum):
            new_unit = tmp_f.unit_x * scalex
            fact = new_unit.asNumber()
            new_unit /= fact
            tmp_f.unit_x = new_unit
            tmp_f.axe_x *= fact
            if fact < 0:
                reversex = True
        else:
            raise TypeError()
        if reversex:
            tmp_f.axe_x = tmp_f.axe_x[::-1]
        # y
        reversey = False
        if scaley is None:
            pass
        elif isinstance(scaley, NUMBERTYPES):
            tmp_f.axe_y *= scaley
            if scaley < 0:
                reversey = True
        elif isinstance(scaley, unum.Unum):
            new_unit = tmp_f.unit_y*scaley
            fact = new_unit.asNumber()
            new_unit /= fact
            tmp_f.unit_y = new_unit
            tmp_f.axe_y *= fact
            if fact < 0:
                reversey = True
        else:
            raise TypeError()
        if reversey:
            tmp_f.axe_y = tmp_f.axe_y[::-1]
        # returning
        if output_reverse:
            if inplace:
                return reversex, reversey
            else:
                return tmp_f, reversex, reversey
        else:
            if inplace:
                pass
            else:
                return tmp_f

    def rotate(self, angle, inplace=False):
        """
        Rotate the field.

        Parameters
        ----------
        angle : integer
            Angle in degrees (positive for trigonometric direction).
            In order to preserve the orthogonal grid, only multiples of
            90° are accepted (can be negative multiples).
        inplace : boolean, optional
            If 'True', Field is rotated in place, else, the function return a
            rotated field.

        Returns
        -------
        rotated_field : Field object, optional
            Rotated field.
        """
        # check params
        if not isinstance(angle, NUMBERTYPES):
            raise TypeError()
        if angle % 90 != 0:
            raise ValueError()
        if not isinstance(inplace, bool):
            raise TypeError()
        # get dat
        if inplace:
            tmp_field = self
        else:
            tmp_field = self.copy()
        # normalize angle
        angle = angle % 360
        # rotate
        if angle == 0:
            pass
        elif angle == 90:
            tmp_field.__axe_x, tmp_field.__axe_y \
                = tmp_field.axe_y[::-1], tmp_field.axe_x
            tmp_field.__unit_x, tmp_field.__unit_y \
                = tmp_field.unit_y, tmp_field.unit_x
        elif angle == 180:
            tmp_field.__axe_x, tmp_field.__axe_y \
                = tmp_field.axe_x[::-1], tmp_field.axe_y[::-1]
        elif angle == 270:
            tmp_field.__axe_x, tmp_field.__axe_y \
                = tmp_field.axe_y, tmp_field.axe_x[::-1]
            tmp_field.__unit_x, tmp_field.__unit_y \
                = tmp_field.unit_y, tmp_field.unit_x
        else:
            raise Exception()
        # correction non-crescent axis
        if tmp_field.axe_x[-1] < tmp_field.axe_x[0]:
            tmp_field.__axe_x = -tmp_field.axe_x
        if tmp_field.axe_y[-1] < tmp_field.axe_y[0]:
            tmp_field.__axe_y = -tmp_field.axe_y
        # returning
        if not inplace:
            return tmp_field

    def change_unit(self, axe, new_unit):
        """
        Change the unit of an Field.

        Parameters
        ----------
        axe : string
            'y' for changing the profile y axis unit
            'x' for changing the profile x axis unit
        new_unit : Unum.unit object or string
            The new unit.
        """
        if isinstance(new_unit, STRINGTYPES):
            new_unit = make_unit(new_unit)
        if not isinstance(new_unit, unum.Unum):
            raise TypeError()
        if not isinstance(axe, STRINGTYPES):
            raise TypeError()
        if axe == 'x':
            old_unit = self.unit_x
            new_unit = old_unit.asUnit(new_unit)
            fact = new_unit.asNumber()
            self.unit_x = new_unit/fact
            self.axe_x *= fact
        elif axe == 'y':
            old_unit = self.unit_y
            new_unit = old_unit.asUnit(new_unit)
            fact = new_unit.asNumber()
            self.unit_y = new_unit/fact
            self.axe_y *= fact
        else:
            raise ValueError()

    def set_origin(self, x=None, y=None):
        """
        Modify the axis in order to place the origin at the givev point (x, y)

        Parameters
        ----------
        x : number
        y : number
        """
        if x is not None:
            if not isinstance(x, NUMBERTYPES):
                raise TypeError("'x' must be a number")
            self.axe_x -= x
        if y is not None:
            if not isinstance(y, NUMBERTYPES):
                raise TypeError("'y' must be a number")
            self.axe_y -= y

    def crop(self, intervx=None, intervy=None, full_output=False,
             ind=False, inplace=False):
        """
        Crop the field in respect with given intervals.

        Parameters
        ----------
        intervx : array, optional
            interval wanted along x
        intervy : array, optional
            interval wanted along y
        full_output : boolean, optional
            If 'True', cutting indices are alson returned
        ind : boolean, optional
            If 'True', intervals are understood as indices along axis.
            If 'False' (default), intervals are understood in axis units.
        inplace : boolean, optional
            If 'True', the field is croped in place.
        """
        # default values
        axe_x, axe_y = self.axe_x, self.axe_y
        if intervx is None:
            if ind:
                intervx = [0, len(axe_x)]
            else:
                intervx = [axe_x[0], axe_x[-1]]
        if intervy is None:
            if ind:
                intervy = [0, len(axe_y)]
            else:
                intervy = [axe_y[0], axe_y[-1]]
        # checking parameters
        if not isinstance(intervx, ARRAYTYPES):
            raise TypeError("'intervx' must be an array of two numbers")
        intervx = np.array(intervx, dtype=float)
        if intervx.ndim != 1:
            raise ValueError("'intervx' must be an array of two numbers")
        if intervx.shape != (2,):
            raise ValueError("'intervx' must be an array of two numbers")
        if intervx[0] > intervx[1]:
            raise ValueError("'intervx' values must be crescent")
        if not isinstance(intervy, ARRAYTYPES):
            raise TypeError("'intervy' must be an array of two numbers")
        intervy = np.array(intervy, dtype=float)
        if intervy.ndim != 1:
            raise ValueError("'intervy' must be an array of two numbers")
        if intervy.shape != (2,):
            raise ValueError("'intervy' must be an array of two numbers")
        if intervy[0] > intervy[1]:
            raise ValueError("'intervy' values must be crescent")
        # checking crooping windows
        if ind:
            if intervx[0] < 0 or intervx[1] == 0 or \
                    intervy[0] < 0 or intervy[1] == 0:
                raise ValueError("Invalid cropping window")
        else:
            if np.all(intervx < axe_x[0]) or np.all(intervx > axe_x[-1])\
                    or np.all(intervy < axe_y[0]) \
                    or np.all(intervy > axe_y[-1]):
                raise ValueError("Invalid cropping window")
        # finding interval indices
        if ind:
            indmin_x = int(intervx[0])
            indmax_x = int(intervx[1])
            indmin_y = int(intervy[0])
            indmax_y = int(intervy[1])
        else:
            if intervx[0] <= axe_x[0]:
                indmin_x = 0
            else:
                indmin_x = self.get_indice_on_axe(1, intervx[0])[-1]
            if intervx[1] >= axe_x[-1]:
                indmax_x = len(axe_x) - 1
            else:
                indmax_x = self.get_indice_on_axe(1, intervx[1])[0]
            if intervy[0] <= axe_y[0]:
                indmin_y = 0
            else:
                indmin_y = self.get_indice_on_axe(2, intervy[0])[-1]
            if intervy[1] >= axe_y[-1]:
                indmax_y = len(axe_y) - 1
            else:
                indmax_y = self.get_indice_on_axe(2, intervy[1])[0]
        # cropping the field
        if inplace:
            axe_x = self.axe_x[indmin_x:indmax_x + 1]
            axe_y = self.axe_y[indmin_y:indmax_y + 1]
            self.__axe_x = axe_x
            self.__axe_y = axe_y
            if full_output:
                return indmin_x, indmax_x, indmin_y, indmax_y
        else:
            cropfield = self.copy()
            cropfield.__axe_x = self.axe_x[indmin_x:indmax_x + 1]
            cropfield.__axe_y = self.axe_y[indmin_y:indmax_y + 1]
            if full_output:
                return indmin_x, indmax_x, indmin_y, indmax_y, cropfield
            else:
                return cropfield

    def extend(self, nmb_left=0, nmb_right=0, nmb_up=0, nmb_down=0,
               inplace=False):
        """
        Add columns or lines of masked values at the field.

        Parameters
        ----------
        nmb_**** : integers
            Number of lines/columns to add in each direction.
        inplace : bool
            If 'False', return a new extended field, if 'True', modify the
            field inplace.
        Returns
        -------
        Extended_field : Field object, optional
            Extended field.
        """
        new_axe_x = self.axe_x.copy()
        new_axe_y = self.axe_y.copy()
        if nmb_left != 0:
            dx = self.axe_x[1] - self.axe_x[0]
            x0 = self.axe_x[0]
            new_xs = np.arange(x0-dx*nmb_left, x0, dx)
            new_axe_x = np.concatenate((new_xs, new_axe_x))
        if nmb_right != 0:
            dx = self.axe_x[-1] - self.axe_x[-2]
            x0 = self.axe_x[-1]
            new_xs = np.arange(x0 + dx, x0+dx*(nmb_right + 1), dx)
            new_axe_x = np.concatenate((new_axe_x, new_xs))
        if nmb_down != 0:
            dy = self.axe_y[1] - self.axe_y[0]
            y0 = self.axe_y[0]
            new_ys = np.arange(y0-dy*nmb_down, y0, dy)
            new_axe_y = np.concatenate((new_ys, new_axe_y))
        if nmb_up != 0:
            dy = self.axe_y[-1] - self.axe_y[-2]
            y0 = self.axe_y[-1]
            new_ys = np.arange(y0 + dy, y0+dy*(nmb_up + 1), dy)
            new_axe_y = np.concatenate((new_axe_y, new_ys))
        if inplace:
            self.__axe_x = new_axe_x
            self.__axe_y = new_axe_y
        else:
            fi = self.copy()
            fi.__axe_x = new_axe_x
            fi.__axe_y = new_axe_y
            return fi

    def __clean(self):
        self.__init__()



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .profile import Profile
from .points import Points
from .temporalpoints import TemporalPoints
from .orientedpoints import OrientedPoints
from .vectorfield import VectorField
from .scalarfield import ScalarField
from .field import Field
from .fields import Fields
from .temporalfields import TemporalFields
from .temporalscalarfields import TemporalScalarFields
from .temporalvectorfields import TemporalVectorFields
from .spatialfields import SpatialFields
from .spatialscalarfields import SpatialScalarFields
from .spatialvectorfields import SpatialVectorFields



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.rst"), "r") as fh:
    long_description = fh.read()

setup(
    name='IMTreatment',
    version='1.3.3',
    description='Tools to analyze scalar and vector fields',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://framagit.org/gabylaunay/IMTreatment',
    author='Gaby Launay',
    author_email='gaby.launay@protonmail.com',
    license='GPLv3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3 :: Only',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    keywords='scalarfield vectorfield fluid dynamics PIV vortex',
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'samples']),
    install_requires=['numpy', 'matplotlib', 'scipy', 'unum', 'imageio',
                      'multiprocess', 'psutil'],
    extras_require={
        'Modal decomposition': 'modred',
        'Davis format support': 'ReadIM',
        'Point clustering': 'sklearn',
        'Mean trajectories clustering': 'networkx',
        'Color in console': 'colorama',
        'Pivmat format support': 'h5py',
        'Export to VTK format': 'pyvtk'},
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-mpl'],
)
